# Autogenerated by Astropy's setup.py on 2020-10-10 15:25:34 UTC
import datetime

version = "4.0.2"
githash = "c20b732a12cd3b06cc639238788207090a34e865"


major = 4
minor = 0
bugfix = 2

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2020, 10, 10, 15, 25, 34)
debug = False

astropy_helpers_version = "4.0.1"
